# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2022 Nexedi SA and Contributors. All Rights Reserved.
#
##############################################################################

from erp5.component.test.SlapOSTestCaseDefaultScenarioMixin import DefaultScenarioMixin
# from DateTime import DateTime

class TestSlapOSVirtualMasterScenario(DefaultScenarioMixin):

  def requestRemoteNode(self, project, remote_project, remote_person):
    remote_node = self.portal.compute_node_module.newContent(
      portal_type='Remote Node',
      title='remote-%s' % self.generateNewId(),
      follow_up_value=project,
      destination_project_value=remote_project.getRelativeUrl(),
      destination_section_value=remote_person.getRelativeUrl(),
      # XXX
      capacity_scope='close'
    )
    self.setServerOpenPublic(remote_node)
    remote_node.setCapacityScope('open')

    # XXX format
    partition = remote_node.newContent(
      portal_type='Compute Partition',
      reference='slapremote0'
    )
    partition.markFree()
    partition.validate()
    remote_node.validate()
    return remote_node

  def addSoftwareProduct(self, title, project, public_server_software,
                         public_instance_type):
    software_product = self.portal.software_product_module.newContent(
      portal_type="Software Product",
      title=title,
      follow_up_value=project,
    )
    software_product.newContent(
      portal_type="Software Product Release Variation",
      title="my old release",
      url_string=public_server_software + '-1'
    )
    release_variation = software_product.newContent(
      portal_type="Software Product Release Variation",
      title="my current release",
      url_string=public_server_software
    )
    software_product.newContent(
      portal_type="Software Product Release Variation",
      title="my futur release",
      url_string=public_server_software + '+1'
    )
    software_product.newContent(
      portal_type="Software Product Type Variation",
      title=public_instance_type + '-1'
    )
    type_variation = software_product.newContent(
      portal_type="Software Product Type Variation",
      title=public_instance_type
    )
    software_product.newContent(
      portal_type="Software Product Type Variation",
      title=public_instance_type + '+1'
    )
    software_product.validate()
    return software_product, release_variation, type_variation


  def stepcheckERP5Consistency(self):
    not_consistent_document = self.portal.portal_catalog.getResultValue(
      consistency_error=1,
      sort_on=[('modification_date', 'DESC')]
    )
    if not_consistent_document is not None:
      assert not_consistent_document.checkConsistency() == [], not_consistent_document.checkConsistency()[0]

  def test_virtual_master_without_accounting_scenario(self):
    # create a default project
    project = self.addProject()
    self.web_site = self.portal.web_site_module.slapos_master_panel

    preference = self.portal.portal_preferences.slapos_default_system_preference
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % project.getRelativeUrl()
      ]
    )

    # some preparation
    self.logout()

    # lets join as slapos administrator, which will own few compute_nodes
    owner_reference = 'owner-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, owner_reference)

    self.login()
    owner_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=owner_reference).getParentValue()

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(owner_person, project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())

    public_server_title = 'Public Server for %s' % owner_reference
    public_server_id = self.requestComputeNode(public_server_title, project.getReference())
    public_server = self.portal.portal_catalog.getResultValue(
        portal_type='Compute Node', reference=public_server_id)
    self.setAccessToMemcached(public_server)
    self.assertNotEqual(None, public_server)
    self.setServerOpenPublic(public_server)

    # and install some software on them
    public_server_software = self.generateNewSoftwareReleaseUrl()
    public_instance_type = 'public type'

    self.supplySoftware(public_server, public_server_software)

    # format the compute_nodes
    self.formatComputeNode(public_server)

    software_product, release_variation, type_variation = self.addSoftwareProduct(
      "instance product", project, public_server_software, public_instance_type
    )

    self.addAllocationSupply("for compute node", public_server, software_product,
                             release_variation, type_variation)

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    public_reference = 'public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, public_reference)

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=public_reference).getParentValue()

    public_instance_title = 'Public title %s' % self.generateNewId()
    self.checkInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        public_server, project.getReference())

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type='ERP5 Login', reference=public_reference).getParentValue()
    self.login(owner_person.getUserId())

    # and the instances
    self.checkInstanceUnallocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type, public_server,
        project.getReference())

    # and uninstall some software on them
    self.logout()
    self.login(owner_person.getUserId())
    self.supplySoftware(public_server, public_server_software,
                        state='destroyed')

    self.logout()
    # Uninstall from compute_node
    self.login()
    self.simulateSlapgridSR(public_server)

    self.tic()

    # XXX Do this for every scenario tests
    self.logout()
    self.stepCallAlarmList()
    self.tic()

    self.login()
    # Ensure no unexpected object has been created
    # 3 (can reduce to 2) assignment, 1 instance tree, 1 compute node, 1 software installation
    # 1 software instance
    # 2 credential request
    # 3 allocation supply, line, cell
    # 1 software product
    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 13, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message


  def test_virtual_master_with_accounting_scenario(self):
    # create a default project
    project = self.addProject()
    self.web_site = self.portal.web_site_module.slapos_master_panel

    preference = self.portal.portal_preferences.slapos_default_system_preference
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % project.getRelativeUrl()
      ]
    )

    public_server_software = self.generateNewSoftwareReleaseUrl()
    public_instance_type = 'public type'

    # XXX who create the product
    # define the security
    organisation = self.portal.organisation_module.newContent(
      portal_type="Organisation"
    )
    # XXX Hardcoded ID
    organisation.newContent(
      portal_type="Bank Account",
      id="bank_account"
    )

    sale_trade_condition = self.portal.sale_trade_condition_module.newContent(
      portal_type="Sale Trade Condition",
      reference='slapos_ultimate_trade_condition',
      # XXX hardcoded
      specialise="business_process_module/slapos_ultimate_business_process",
      destination_project_value=project,
      source_value=organisation,
      source_section_value=organisation,
      price_currency="currency_module/EUR"
    )
    sale_trade_condition.newContent(
      portal_type="Trade Model Line",
      reference="VAT",
      resource="service_module/slapos_tax",
      base_application="base_amount/invoicing/taxable",
      trade_phase="slapos/tax",
      price=0.2,
      quantity=1.0,
      membership_criterion_base_category=('price_currency',),
      membership_criterion_category=('price_currency/currency_module/EUR',)
    )

    sale_trade_condition.validate()

    software_product, release_variation, type_variation = self.addSoftwareProduct(
      "instance product", project, public_server_software, public_instance_type
    )

    sale_supply = self.portal.sale_supply_module.newContent(
      portal_type="Sale Supply",
      destination_project_value=project,
      price_currency="currency_module/EUR"
    )
    # XXX Put price in sale supply module
    sale_supply.newContent(
      portal_type="Sale Supply Line",
      base_price=9,
      resource_value=software_product
    )
    sale_supply.validate()

    # some preparation
    self.logout()

    # lets join as slapos administrator, which will own few compute_nodes
    owner_reference = 'owner-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, owner_reference)

    self.login()
    owner_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=owner_reference).getParentValue()

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(owner_person, project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())

    public_server_title = 'Public Server for %s' % owner_reference
    public_server_id = self.requestComputeNode(public_server_title, project.getReference())
    public_server = self.portal.portal_catalog.getResultValue(
        portal_type='Compute Node', reference=public_server_id)
    self.setAccessToMemcached(public_server)
    self.assertNotEqual(None, public_server)
    self.setServerOpenPublic(public_server)

    self.addAllocationSupply("for compute node", public_server, software_product,
                             release_variation, type_variation)

    # and install some software on them
    self.supplySoftware(public_server, public_server_software)

    # format the compute_nodes
    self.formatComputeNode(public_server)

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    public_reference = 'public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, public_reference)

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=public_reference).getParentValue()

    public_instance_title = 'Public title %s' % self.generateNewId()
    self.checkInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        public_server, project.getReference())

    # XXX Do this for every scenario tests
    self.logout()
    for _ in range(20):
      self.stepCallAlarmList()
      self.tic()

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type='ERP5 Login', reference=public_reference).getParentValue()
    self.login(owner_person.getUserId())

    # and the instances
    self.checkInstanceUnallocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type, public_server,
        project.getReference())

    # and uninstall some software on them
    self.logout()
    self.login(owner_person.getUserId())
    self.supplySoftware(public_server, public_server_software,
                        state='destroyed')

    self.logout()
    # Uninstall from compute_node
    self.login()
    self.simulateSlapgridSR(public_server)

    self.tic()

    # XXX Do this for every scenario tests
    # XXX Do this for every scenario tests
    self.logout()
    for _ in range(20):
      self.stepCallAlarmList()
      self.tic()
    self.login()

    # Check stock
    inventory_list = self.portal.portal_simulation.getCurrentInventoryList(**{
      'group_by_section': False,
      'group_by_node': True,
      'group_by_variation': True,
      'resource_uid': software_product.getUid(),
      'node_uid': public_person.getUid(),
      'project_uid': project.getUid(),
      'ledger_uid': self.portal.portal_categories.ledger.automated.getUid()
    })
    assert len(inventory_list) == 1, len(inventory_list)
    assert inventory_list[0].quantity == 1, inventory_list[0].quantity
    resource_vcl = [
      'software_release/%s' % release_variation.getRelativeUrl(),
      'software_type/%s' % type_variation.getRelativeUrl()
    ]
    resource_vcl.sort()
    assert inventory_list[0].getVariationCategoryList() == resource_vcl, "%s %s" % (resource_vcl, inventory_list[0].getVariationCategoryList())

    # Check accounting
    transaction_list = self.portal.account_module.receivable.Account_getAccountingTransactionList(mirror_section_uid=public_person.getUid())
    assert len(transaction_list) == 1, len(transaction_list)
    assert transaction_list[0].total_price == 10.8, transaction_list[0].total_price

    self.login()

    # Ensure no unexpected object has been created
    # 3 (can reduce to 2) assignment, 1 instance tree, 1 compute node, 1 software installation
    # 1 software instance
    # 2 credential request

    # 1 open order
    # 1 hosting subscription
    # 2 accounting transaction
    # 1 packing list
    # 7 simulation mvt

    # 1 software product
    # 1 sale trace condition
    # 1 supply, 1 supply line
    # 1 open order line / 1 cell
    # 3 allocation supply / line / cell

    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 30, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message


  def addInstanceNode(self, title, software_instance):
    instance_node = self.portal.compute_node_module.newContent(
      portal_type='Instance Node',
      title=title,
      specialise_value=software_instance,
      follow_up_value=software_instance.getFollowUpValue()
    )
    instance_node.validate()
    return instance_node


  def test_virtual_master_slave_without_accounting_scenario(self):
    # create a default project
    project = self.addProject()
    self.web_site = self.portal.web_site_module.slapos_master_panel

    preference = self.portal.portal_preferences.slapos_default_system_preference
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % project.getRelativeUrl()
      ]
    )

    # some preparation
    self.logout()

    # lets join as slapos administrator, which will own few compute_nodes
    owner_reference = 'owner-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, owner_reference)

    self.login()
    owner_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=owner_reference).getParentValue()

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(owner_person, project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())

    public_server_title = 'Public Server for %s' % owner_reference
    public_server_id = self.requestComputeNode(public_server_title, project.getReference())
    public_server = self.portal.portal_catalog.getResultValue(
        portal_type='Compute Node', reference=public_server_id)
    self.setAccessToMemcached(public_server)
    self.assertNotEqual(None, public_server)
    self.setServerOpenPublic(public_server)

    # and install some software on them
    public_server_software = self.generateNewSoftwareReleaseUrl()
    self.supplySoftware(public_server, public_server_software)

    #software_product, release_variation, type_variation = self.addSoftwareProduct(
    public_instance_type = 'public type'
    software_product, software_release, software_type = self.addSoftwareProduct(
      "instance product", project, public_server_software, public_instance_type
    )

    self.addAllocationSupply("for compute node", public_server, software_product,
                             software_release, software_type)

    # format the compute_nodes
    self.formatComputeNode(public_server)

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    public_reference = 'public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, public_reference)

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=public_reference).getParentValue()

    public_instance_title = 'Public title %s' % self.generateNewId()
    self.checkInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        public_server, project.getReference())

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())
    instance_node_title = 'Shared Instance for %s' % owner_reference
    # Convert the Software Instance into an Instance Node
    # to explicitely mark it as accepting Slave Instance
    software_instance = self.portal.portal_catalog.getResultValue(
        portal_type='Software Instance', title=public_instance_title)
    instance_node = self.addInstanceNode(instance_node_title, software_instance)

    slave_server_software = self.generateNewSoftwareReleaseUrl()
    slave_instance_type = 'slave type'
    software_product, software_release, software_type = self.addSoftwareProduct(
      'share product', project, slave_server_software, slave_instance_type
    )
    self.addAllocationSupply("for instance node", instance_node, software_product,
                             software_release, software_type)

    slave_instance_title = 'Slave title %s' % self.generateNewId()
    self.checkSlaveInstanceAllocation(public_person.getUserId(),
        public_reference, slave_instance_title,
        slave_server_software, slave_instance_type,
        public_server, project.getReference())

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type='ERP5 Login', reference=public_reference).getParentValue()
    self.login(owner_person.getUserId())

    # and the instances
    self.checkSlaveInstanceUnallocation(public_person.getUserId(),
        public_reference, slave_instance_title,
        slave_server_software, slave_instance_type, public_server,
        project.getReference())

    # and uninstall some software on them
    self.logout()
    self.login(owner_person.getUserId())
    self.supplySoftware(public_server, public_server_software,
                        state='destroyed')

    self.logout()
    # Uninstall from compute_node
    self.login()
    self.simulateSlapgridSR(public_server)

    self.tic()

    # XXX Do this for every scenario tests
    self.logout()
    self.stepCallAlarmList()
    self.tic()

    self.login()
    # Ensure no unexpected object has been created
    # 3 (can reduce to 2) assignment, 2 instance tree, 1 compute node, 1 software installation
    # 2 software instance
    # 2 credential request
    # 2 software product
    # 6 allocation supply/line/cell
    # 1 instance node
    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 20, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message


  def test_virtual_master_slave_on_same_tree_without_accounting_scenario(self):
    # create a default project
    project = self.addProject()
    self.web_site = self.portal.web_site_module.slapos_master_panel

    preference = self.portal.portal_preferences.slapos_default_system_preference
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % project.getRelativeUrl()
      ]
    )

    # some preparation
    self.logout()

    # lets join as slapos administrator, which will own few compute_nodes
    owner_reference = 'owner-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, owner_reference)

    self.login()
    owner_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=owner_reference).getParentValue()

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(owner_person, project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())

    public_server_title = 'Public Server for %s' % owner_reference
    public_server_id = self.requestComputeNode(public_server_title, project.getReference())
    public_server = self.portal.portal_catalog.getResultValue(
        portal_type='Compute Node', reference=public_server_id)
    self.setAccessToMemcached(public_server)
    self.assertNotEqual(None, public_server)
    self.setServerOpenPublic(public_server)

    # and install some software on them
    public_server_software = self.generateNewSoftwareReleaseUrl()
    self.supplySoftware(public_server, public_server_software)

    #software_product, release_variation, type_variation = self.addSoftwareProduct(
    public_instance_type = 'public type'
    software_product, software_release, software_type = self.addSoftwareProduct(
      "instance product", project, public_server_software, public_instance_type
    )

    self.addAllocationSupply("for compute node", public_server, software_product,
                             software_release, software_type,
                             is_slave_on_same_instance_tree_allocable=True)

    # format the compute_nodes
    self.formatComputeNode(public_server)

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    public_reference = 'public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, public_reference)

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=public_reference).getParentValue()

    public_instance_title = 'Public title %s' % self.generateNewId()
    self.checkInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        public_server, project.getReference())

    self.tic()

    # request slave instance on the same instance tree
    self.login(public_person.getUserId())
    slave_instance_title = 'Slave title %s' % self.generateNewId()
    self.checkInstanceTreeSlaveInstanceAllocation(
      public_person.getUserId(),
      public_reference, public_instance_title,
      slave_instance_title,
      public_server_software, public_instance_type,
      public_server, project.getReference()
    )

    self.login()
    # Ensure no unexpected object has been created
    # 3 (can reduce to 2) assignment, 1 instance tree, 1 compute node, 1 software installation
    # 2 software instance
    # 2 credential request
    # 1 software product
    # 3 allocation supply/line/cell
    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 14, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message


  def test_virtual_master_on_remote_tree_without_accounting_scenario(self):
    self.web_site = self.portal.web_site_module.slapos_master_panel
    preference = self.portal.portal_preferences.slapos_default_system_preference

    ####################################
    # Create a remote project
    ####################################
    remote_project = self.addProject()
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % remote_project.getRelativeUrl()
      ]
    )

    # some preparation
    self.logout()

    # lets join as slapos administrator, which will own few compute_nodes
    remote_owner_reference = 'remote-owner-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, remote_owner_reference)

    self.login()
    remote_owner_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=remote_owner_reference).getParentValue()

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(remote_owner_person, remote_project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(remote_owner_person.getUserId())

    remote_server_title = 'Remote Server for %s' % remote_owner_person
    remote_server_id = self.requestComputeNode(remote_server_title, remote_project.getReference())
    remote_server = self.portal.portal_catalog.getResultValue(
        portal_type='Compute Node', reference=remote_server_id)
    self.setAccessToMemcached(remote_server)
    self.assertNotEqual(None, remote_server)
    self.setServerOpenPublic(remote_server)

    # and install some software on them
    remote_server_software = self.generateNewSoftwareReleaseUrl()
    remote_instance_type = 'public type'

    self.supplySoftware(remote_server, remote_server_software)

    # format the compute_nodes
    self.formatComputeNode(remote_server)

    remote_software_product, remote_release_variation, remote_type_variation = self.addSoftwareProduct(
      "remote product", remote_project, remote_server_software, remote_instance_type
    )

    self.addAllocationSupply("for compute node", remote_server, remote_software_product,
                             remote_release_variation, remote_type_variation)

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    remote_public_reference = 'remote-public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, remote_public_reference)

    self.login()
    remote_public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=remote_public_reference).getParentValue()

    ####################################
    # Create a local project
    ####################################
    project = self.addProject()
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % project.getRelativeUrl()
      ]
    )

    owner_person = remote_public_person

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(owner_person, project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())

    remote_compute_node = self.requestRemoteNode(project, remote_project,
                                           remote_public_person)

    # and install some software on them
    public_server_software = remote_server_software

    #software_product, release_variation, type_variation = self.addSoftwareProduct(
    public_instance_type = remote_instance_type
    software_product, software_release, software_type = self.addSoftwareProduct(
      "instance product", project, public_server_software, public_instance_type
    )

    self.addAllocationSupply("for remote node", remote_compute_node, software_product,
                             software_release, software_type)
    self.tic()

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    public_reference = 'public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, public_reference)

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=public_reference).getParentValue()

    public_instance_title = 'Public title %s' % self.generateNewId()
    self.checkRemoteInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        remote_compute_node, project.getReference())

    # XXX Do this for every scenario tests
    self.logout()
    for _ in range(3):
      self.stepCallAlarmList()
      self.tic()
      # now instantiate it on compute_node and set some nice connection dict
      self.simulateSlapgridCP(remote_server)
      self.tic()
    self.login()

    # owner_person should have one Instance Tree created by alarm
    owner_instance_tree_list = self.portal.portal_catalog(
      portal_type='Instance Tree',
      destination_section__uid=owner_person.getUid()
    )
    self.assertEqual(1, len(owner_instance_tree_list))
    owner_software_instance = owner_instance_tree_list[0].getSuccessorValue()
    self.assertEqual('Software Instance', owner_software_instance.getPortalType())
    self.assertEqual(
      remote_server.getRelativeUrl(),
      owner_software_instance.getAggregateValue().getParentValue().getRelativeUrl()
    )

    # public_person should have one Instance Tree
    public_instance_tree_list = self.portal.portal_catalog(
      portal_type='Instance Tree',
      destination_section__uid=public_person.getUid()
    )
    self.assertEqual(1, len(public_instance_tree_list))

    self.assertEqual(
      '_remote_%s_%s' % (project.getReference(),
                         public_instance_tree_list[0].getSuccessorReference()),
      owner_software_instance.getTitle()
    )
    connection_dict = owner_software_instance.getConnectionXmlAsDict()
    self.assertSameSet(('url_1', 'url_2'), connection_dict.keys())
    self.assertSameSet(
        ['http://%s/' % q.getIpAddress() for q in
            owner_software_instance.getAggregateValue().contentValues(portal_type='Internet Protocol Address')],
        connection_dict.values())

    self.checkRemoteInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        remote_compute_node, project.getReference(),
        connection_dict_to_check=owner_software_instance.getConnectionXmlAsDict())

    self.login()

    # Ensure no unexpected object has been created
    # 3 (can reduce to 2) assignment, 1 remote node, 1 software installation
    # 1 compute node
    # 1 instance tree, 1 software instance
    # 2 credential request
    # 1 software product
    # 3 allocation supply/line/cell
    related_object_list = remote_project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 14, [x.getRelativeUrl() for x in related_object_list]

    # Ensure no unexpected object has been created
    # 2 (can reduce to 2) assignment, 1 remote node
    # 1 instance tree, 1 software instance
    # 1 credential request
    # 1 software product
    # 3 allocation supply/line/cell
    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 10, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message



  def test_virtual_master_slave_instance_on_remote_tree_without_accounting_scenario(self):
    self.web_site = self.portal.web_site_module.slapos_master_panel
    preference = self.portal.portal_preferences.slapos_default_system_preference

    ####################################
    # Create a remote project
    ####################################
    remote_project = self.addProject()
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % remote_project.getRelativeUrl()
      ]
    )

    # some preparation
    self.logout()

    # lets join as slapos administrator, which will own few compute_nodes
    remote_owner_reference = 'remote-owner-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, remote_owner_reference)

    self.login()
    remote_owner_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=remote_owner_reference).getParentValue()

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(remote_owner_person, remote_project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(remote_owner_person.getUserId())

    remote_server_title = 'Remote Server for %s' % remote_owner_person
    remote_server_id = self.requestComputeNode(remote_server_title, remote_project.getReference())
    remote_server = self.portal.portal_catalog.getResultValue(
        portal_type='Compute Node', reference=remote_server_id)
    self.setAccessToMemcached(remote_server)
    self.assertNotEqual(None, remote_server)
    self.setServerOpenPublic(remote_server)

    # and install some software on them
    remote_server_software = self.generateNewSoftwareReleaseUrl()
    remote_instance_type = 'public type'

    self.supplySoftware(remote_server, remote_server_software)

    # format the compute_nodes
    self.formatComputeNode(remote_server)

    remote_software_product, remote_release_variation, remote_type_variation = self.addSoftwareProduct(
      "remote product", remote_project, remote_server_software, remote_instance_type
    )

    self.addAllocationSupply("for compute node", remote_server, remote_software_product,
                             remote_release_variation, remote_type_variation,
                             destination_value=remote_owner_person)

    private_instance_title = 'Private title %s' % self.generateNewId()
    self.checkInstanceAllocation(remote_owner_person.getUserId(),
        remote_owner_reference, private_instance_title,
        remote_server_software, remote_instance_type,
        remote_server, remote_project.getReference())

    # Convert the Software Instance into an Instance Node
    # to explicitely mark it as accepting Slave Instance
    private_software_instance = self.portal.portal_catalog.getResultValue(
        portal_type='Software Instance', title=private_instance_title)
    instance_node_title = 'Shared Instance for %s' % remote_owner_reference
    instance_node = self.addInstanceNode(instance_node_title, private_software_instance)

    self.addAllocationSupply(
      "for instance node", instance_node, remote_software_product,
      remote_release_variation, remote_type_variation)

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    remote_public_reference = 'remote-public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, remote_public_reference)

    self.login()
    remote_public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=remote_public_reference).getParentValue()

    ####################################
    # Create a local project
    ####################################
    project = self.addProject()
    preference.edit(
      preferred_subscription_assignment_category_list=[
        'function/customer',
        'role/client',
        'destination_project/%s' % project.getRelativeUrl()
      ]
    )

    owner_person = remote_public_person

    # first slapos administrator assignment can only be created by
    # the erp5 manager
    self.addProjectProductionManagerAssignment(owner_person, project)
    self.tic()

    # hooray, now it is time to create compute_nodes
    self.login(owner_person.getUserId())

    remote_compute_node = self.requestRemoteNode(project, remote_project,
                                           remote_public_person)

    # and install some software on them
    public_server_software = remote_server_software

    remote_compute_node.requestSoftwareRelease(software_release_url=public_server_software,
                                         state='available')

    #software_product, release_variation, type_variation = self.addSoftwareProduct(
    public_instance_type = remote_instance_type
    software_product, software_release, software_type = self.addSoftwareProduct(
      "instance product", project, public_server_software, public_instance_type
    )

    self.addAllocationSupply("for remote node", remote_compute_node, software_product,
                             software_release, software_type)
    self.tic()

    # join as the another visitor and request software instance on public
    # compute_node
    self.logout()
    public_reference = 'public-%s' % self.generateNewId()
    self.joinSlapOS(self.web_site, public_reference)

    self.login()
    public_person = self.portal.portal_catalog.getResultValue(
      portal_type="ERP5 Login",
      reference=public_reference).getParentValue()

    public_instance_title = 'Public title %s' % self.generateNewId()
    self.checkRemoteInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        remote_compute_node, project.getReference(),
        slave=True)

    # XXX Do this for every scenario tests
    self.logout()
    for _ in range(3):
      self.stepCallAlarmList()
      self.tic()
      # now instantiate it on compute_node and set some nice connection dict
      self.simulateSlapgridCP(remote_server)
      self.tic()
    self.login()

    # owner_person should have one Instance Tree created by alarm
    owner_instance_tree_list = self.portal.portal_catalog(
      portal_type='Instance Tree',
      destination_section__uid=owner_person.getUid()
    )
    self.assertEqual(1, len(owner_instance_tree_list))
    owner_software_instance = owner_instance_tree_list[0].getSuccessorValue()
    self.assertEqual('Slave Instance', owner_software_instance.getPortalType())
    self.assertEqual(
      remote_server.getRelativeUrl(),
      owner_software_instance.getAggregateValue().getParentValue().getRelativeUrl()
    )

    # public_person should have one Instance Tree
    public_instance_tree_list = self.portal.portal_catalog(
      portal_type='Instance Tree',
      destination_section__uid=public_person.getUid()
    )
    self.assertEqual(1, len(public_instance_tree_list))

    self.assertEqual(
      '_remote_%s_%s' % (project.getReference(),
                         public_instance_tree_list[0].getSuccessorReference()),
      owner_software_instance.getTitle()
    )
    connection_dict = owner_software_instance.getConnectionXmlAsDict()
    self.assertSameSet(('url_1', 'url_2'), connection_dict.keys())
    self.assertSameSet(
        ['http://%s/%s' % (q.getIpAddress(), owner_software_instance.getReference()) for q in
            owner_software_instance.getAggregateValue().contentValues(portal_type='Internet Protocol Address')],
        connection_dict.values())

    self.checkRemoteInstanceAllocation(public_person.getUserId(),
        public_reference, public_instance_title,
        public_server_software, public_instance_type,
        remote_compute_node, project.getReference(),
        connection_dict_to_check=owner_software_instance.getConnectionXmlAsDict(),
        slave=True)

    self.login()

    # Ensure no unexpected object has been created
    # 3 (can reduce to 2) assignment, 1 remote node, 1 software installation
    # 1 compute node
    # 1 instance node
    # 2 instance tree, 1 software instance, 1 slave instance
    # 2 credential request
    # 1 software product
    # 6 allocation supply/line/cell
    related_object_list = remote_project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 20, [x.getRelativeUrl() for x in related_object_list]

    # Ensure no unexpected object has been created
    # 2 (can reduce to 2) assignment, 1 remote node, 1 software installation
    # 1 instance tree, 1 slave instance
    # 1 credential request
    # 1 software product
    # 3 allocation supply/line/cell
    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 11, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message


  def test_open_order_with_service_scenario(self):

    # create a default project
    project = self.addProject()
    person = self.portal.person_module.newContent(
      portal_type="Person",
      default_email_coordinate_text='a@example.org',
    )
    organisation = self.portal.organisation_module.newContent(
      portal_type="Organisation"
    )

    sale_trade_condition = self.portal.sale_trade_condition_module.newContent(
      portal_type="Sale Trade Condition",
      reference='slapos_ultimate_trade_condition',
      # XXX hardcoded
      specialise="business_process_module/slapos_ultimate_business_process"
    )
    sale_trade_condition.newContent(
      portal_type="Trade Model Line",
      reference="VAT",
      resource="service_module/slapos_tax",
      base_application="base_amount/invoicing/taxable",
      trade_phase="slapos/tax",
      price=0.2,
      quantity=1.0,
      membership_criterion_base_category=('price_currency',),
      membership_criterion_category=('price_currency/currency_module/EUR',)
    )
    sale_trade_condition.validate()

    service = self.portal.restrictedTraverse('service_module/slapos_virtual_master_subscription')

    for _ in range(1):
      hosting_subscription = self.portal.hosting_subscription_module.newContent(
        portal_type="Hosting Subscription",
        # XXX hardcoded
        ledger="automated",
        follow_up_value=project,
      )
      hosting_subscription.validate()
      start_date = hosting_subscription.HostingSubscription_calculateSubscriptionStartDate()

      # create open order
      open_sale_order = self.portal.open_sale_order_module.newContent(
        portal_type="Open Sale Order",
        ledger=hosting_subscription.getLedger(),
        destination_project_value=project,
        source_value=organisation,
        source_section_value=organisation,
        destination_value=person,
        destination_section_value=person,
        destination_decision_value=person,
        price_currency='currency_module/EUR',
        payment_mode="payzen",
        start_date=start_date,
        # Ensure stop date value is higher than start date
        # it will be updated by OpenSaleOrder_updatePeriod
        stop_date=start_date + 1,
        specialise_value=sale_trade_condition
      )

      open_sale_order.newContent(
        portal_type="Open Sale Order Line",
        quantity=10,
        price=2,
        resource_value=service,
        aggregate_value=[
          hosting_subscription,
          project
        ]
      )
      open_sale_order.order()
      open_sale_order.validate()
    self.tic()

    # XXX Do this for every scenario tests
    self.logout()
    for _ in range(20):
      self.stepCallAlarmList()
      self.tic()
    self.login()

    # Check stock
    inventory_list = self.portal.portal_simulation.getCurrentInventoryList(**{
      'group_by_section': False,
      'group_by_node': True,
      'group_by_variation': True,
      'resource_uid': service.getUid(),
      'node_uid': person.getUid(),
      'project_uid': project.getUid(),
      'ledger_uid': hosting_subscription.getLedgerUid()
    })
    assert len(inventory_list) == 1, len(inventory_list)
    assert inventory_list[0].quantity == 10, inventory_list[0].quantity
    assert inventory_list[0].getVariationCategoryList() == [], inventory_list[0].getVariationCategoryList()

    # Check accounting
    transaction_list = self.portal.account_module.receivable.Account_getAccountingTransactionList(mirror_section_uid=person.getUid())
    assert len(transaction_list) == 1, len(transaction_list)
    assert transaction_list[0].total_price == 24, transaction_list[0].total_price

    # Ensure no unexpected object has been created
    # destination project:
    # 1 open order
    # 1 hosting subscription
    # 2 accounting transaction
    # 1 packing list
    # 7 simulation mvt
    # aggregate:
    # 1 invoice line
    # 1 packing list line
    # 1 open order line

    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 15, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message


  def test_open_order_with_software_product_scenario(self):
    # create a default project
    project = self.addProject()
    person = self.portal.person_module.newContent(
      portal_type="Person",
      default_email_coordinate_text='a@example.org',
    )
    organisation = self.portal.organisation_module.newContent(
      portal_type="Organisation"
    )

    sale_trade_condition = self.portal.sale_trade_condition_module.newContent(
      portal_type="Sale Trade Condition",
      reference='slapos_ultimate_trade_condition',
      # XXX hardcoded
      specialise="business_process_module/slapos_ultimate_business_process"
    )
    sale_trade_condition.newContent(
      portal_type="Trade Model Line",
      reference="VAT",
      resource="service_module/slapos_tax",
      base_application="base_amount/invoicing/taxable",
      trade_phase="slapos/tax",
      price=0.2,
      quantity=1.0,
      membership_criterion_base_category=('price_currency',),
      membership_criterion_category=('price_currency/currency_module/EUR',)
    )
    sale_trade_condition.validate()

    software_product = self.portal.software_product_module.newContent(
      portal_type="Software Product",
      title="foo software product",
      follow_up_value=project,
    )
    software_product.newContent(
      portal_type="Software Product Release Variation",
      title="my super release"
    )
    software_product.newContent(
      portal_type="Software Product Type Variation",
      title="my super type"
    )
    software_product.validate()

    instance_tree = self.portal.instance_tree_module.newContent(
      portal_type="Instance Tree",
      follow_up_value=project
    )

    for _ in range(1):
      hosting_subscription = self.portal.hosting_subscription_module.newContent(
        portal_type="Hosting Subscription",
        # XXX hardcoded
        ledger="automated",
      )
      hosting_subscription.validate()
      start_date = hosting_subscription.HostingSubscription_calculateSubscriptionStartDate()

      # create open order
      open_sale_order = self.portal.open_sale_order_module.newContent(
        portal_type="Open Sale Order",
        ledger=hosting_subscription.getLedger(),
        destination_project_value=project,
        source_value=organisation,
        source_section_value=organisation,
        destination_value=person,
        destination_section_value=person,
        destination_decision_value=person,
        price_currency='currency_module/EUR',
        payment_mode="payzen",
        start_date=start_date,
        # Ensure stop date value is higher than start date
        # it will be updated by OpenSaleOrder_updatePeriod
        stop_date=start_date + 1,
        specialise_value=sale_trade_condition
      )

      resource_vcl = list(software_product.getVariationCategoryList(
                                     omit_individual_variation=0))
      resource_vcl.sort()
      open_order_line = open_sale_order.newContent(
        portal_type="Open Sale Order Line",
        resource_value=software_product,
        variation_category_list=resource_vcl
      )


      base_id = 'path'
      cell_key = list(open_order_line.getCellKeyList(base_id=base_id))[0]
      #cell_key_list.sort()
      open_order_cell = open_order_line.newCell(
        base_id=base_id,
        portal_type='Open Sale Order Cell',
        *cell_key
      )
      open_order_cell.edit(
        mapped_value_property_list=['price','quantity'],
        price=3,
        quantity=4,
        predicate_category_list=cell_key,
        variation_category_list=cell_key,
        aggregate_value=[
          hosting_subscription,
          instance_tree
        ],
      )

      open_sale_order.order()
      open_sale_order.validate()
    self.tic()

    # XXX Do this for every scenario tests
    self.logout()
    for _ in range(20):
      self.stepCallAlarmList()
      self.tic()
    self.login()

    # Check stock
    inventory_list = self.portal.portal_simulation.getCurrentInventoryList(**{
      'group_by_section': False,
      'group_by_node': True,
      'group_by_variation': True,
      'resource_uid': software_product.getUid(),
      'node_uid': person.getUid(),
      'project_uid': project.getUid(),
      'ledger_uid': hosting_subscription.getLedgerUid()
    })
    assert len(inventory_list) == 1, len(inventory_list)
    assert inventory_list[0].quantity == 4, inventory_list[0].quantity
    assert inventory_list[0].getVariationCategoryList() == resource_vcl, inventory_list[0].getVariationCategoryList()

    # Check accounting
    transaction_list = self.portal.account_module.receivable.Account_getAccountingTransactionList(mirror_section_uid=person.getUid())
    assert len(transaction_list) == 1, len(transaction_list)
    assert transaction_list[0].total_price == 14.4, transaction_list[0].total_price

    # Ensure no unexpected object has been created
    # destination project:
    # 1 open order
    # 2 accounting transaction
    # 1 packing list
    # 7 simulation mvt
    # 1 instance tree
    # 1 software product
    # acquisition...
    # 1 open order line
    # 1 open order cell

    related_object_list = project.Base_getRelatedObjectList(**{'category.category_strict_membership': 1})
    assert len(related_object_list) == 15, [x.getRelativeUrl() for x in related_object_list]

    self.stepcheckERP5Consistency()

    # after accept, an email is send containing the reset link
    last_message = self.portal.MailHost._last_message
    assert last_message is None, last_message

