from Products.ERP5Security import SUPER_USER
from AccessControl.SecurityManagement import getSecurityManager
from AccessControl.SecurityManagement import setSecurityManager
from AccessControl.SecurityManagement import newSecurityManager


def ERP5Site_activateAlarmSlapOSPanelTest(self):
  portal = self.getPortalObject()
  sm = getSecurityManager()
  try:
    newSecurityManager(None, portal.acl_users.getUser(SUPER_USER))

    for alarm in portal.portal_alarms.contentValues():
      if alarm.isEnabled():
        alarm.activeSense()

  finally:
    setSecurityManager(sm)

  return "Alarms activated."


def ERP5Site_bootstrapSlapOSPanelTest(self, scenario, customer_login,
                                      manager_login, remote_customer_login,
                                      passwd):

  if scenario not in ['accounting', 'customer', 'customer_shared',
                      'customer_remote']:
    raise ValueError('Unsupported bootstrap scenario: %s' % scenario)

  portal = self.getPortalObject()
  sm = getSecurityManager()
  try:
    newSecurityManager(None, portal.acl_users.getUser(SUPER_USER))

    if scenario == 'accounting':
      organisation = portal.organisation_module.newContent(
        portal_type="Organisation",
        title="Test accounting organisation"
      )
    else:
      organisation = None

    # Create Project
    project = portal.project_module.newContent(
      portal_type="Project",
      title='Test Project',
      destination_section_value=organisation
    )
    project.validate()

    if scenario == 'customer_remote':
      remote_project = portal.project_module.newContent(
        portal_type="Project",
        title='Test Remote Project'
      )
      remote_project.validate()

    # Bootstrap one manager user
    manager_person = portal.person_module.newContent(
      portal_type='Person',
      first_name='Manual test Project Manager',
      default_email_coordinate_text='romain+manager@nexedi.com'
    )
    manager_person.newContent(
      portal_type='Assignment',
      title='Production Manager for project %s' % project.getTitle(),
      destination_project_value=project,
      function='production/manager'
    ).open()
    if scenario == 'customer_shared':
      # XXX For shared instance, user must also be a customer
      # How to create Instance Node without any user related document?
      manager_person.newContent(
        portal_type='Assignment',
        title='Customer for project %s' % project.getTitle(),
        destination_project_value=project,
        function='customer'
      ).open()
    manager_person.newContent(
      portal_type='ERP5 Login',
      reference=manager_login,
      password=passwd
    ).validate()
    manager_person.validate()

    # Bootstrap one customer user
    customer_person = portal.person_module.newContent(
      portal_type='Person',
      first_name='Manual test Project Customer',
      default_email_coordinate_text='romain+customer@nexedi.com'
    )
    customer_person.newContent(
      portal_type='Assignment',
      title='Customer for project %s' % project.getTitle(),
      destination_project_value=project,
      function='customer'
    ).open()
    customer_person.newContent(
      portal_type='ERP5 Login',
      reference=customer_login,
      password=passwd
    ).validate()
    customer_person.validate()

    if scenario == 'customer_remote':
      customer_person.newContent(
        portal_type='Assignment',
        title='Production Manager for project %s' % remote_project.getTitle(),
        destination_project_value=remote_project,
        function='production/manager'
      ).open()

      # Bootstrap one customer user for the remote project
      remote_customer_person = portal.person_module.newContent(
        portal_type='Person',
        first_name='Manual test Remote Project Customer',
        default_email_coordinate_text='romain+remote+customer@nexedi.com'
      )
      remote_customer_person.newContent(
        portal_type='Assignment',
        title='Customer for project %s' % remote_project.getTitle(),
        destination_project_value=remote_project,
        function='customer'
      ).open()
      remote_customer_person.newContent(
        portal_type='ERP5 Login',
        reference=remote_customer_login,
        password=passwd
      ).validate()
      remote_customer_person.validate()

  finally:
    setSecurityManager(sm)

  return "Done."

