# -*- coding:utf-8 -*-
##############################################################################
#
# Copyright (c) 2002-2013 Nexedi SA and Contributors. All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
#
##############################################################################
from erp5.component.test.SlapOSTestCaseMixin import SlapOSTestCaseMixin, TemporaryAlarmScript

class TestSlapOSUpgradeDecisionProcess(SlapOSTestCaseMixin):

  def _makeUpgradeDecision(self, confirm=True):
    upgrade_decision = self.portal.\
       upgrade_decision_module.newContent(
         portal_type="Upgrade Decision",
         title="TESTUPDE-%s" % self.generateNewId())
    if confirm:
      upgrade_decision.confirm()
    return upgrade_decision

  def _makeInstanceTree(self, slap_state="start_requested"):
    instance_tree = self.portal\
      .instance_tree_module.template_instance_tree\
      .Base_createCloneDocument(batch_mode=1)
    instance_tree.validate()
    new_id = self.generateNewId()
    instance_tree.edit(
        title= "Test hosting sub start %s" % new_id,
        reference="TESTHSS-%s" % new_id
    )
    self.portal.portal_workflow._jumpToStateFor(instance_tree, slap_state)

    return instance_tree

  def test_alarm_upgrade_decision_process_instance_tree(self):
    upgrade_decision = self._makeUpgradeDecision()
    upgrade_decision.start()
    self.tic()

    with TemporaryAlarmScript(self.portal, 'UpgradeDecision_processUpgrade'):
      self.portal.portal_alarms.slapos_pdm_upgrade_decision_process_started.\
        activeSense()
      self.tic()

    self.assertEqual(
        'Visited by UpgradeDecision_processUpgrade',
        upgrade_decision.workflow_history['edit_workflow'][-1]['comment'])

  def test_alarm_instance_tree_create_upgrade_decision(self):
    instance_tree = self._makeInstanceTree()
    self.tic()

    with TemporaryAlarmScript(self.portal, 'InstanceTree_createUpgradeDecision'):
      self.portal.portal_alarms.slapos_pdm_instance_tree_create_upgrade_decision.\
        activeSense()
      self.tic()

    self.assertEqual('Visited by InstanceTree_createUpgradeDecision',
      instance_tree.workflow_history['edit_workflow'][-1]['comment'])

  def test_alarm_create_upgrade_decision_destroyed_instance_tree(self):
    instance_tree = self._makeInstanceTree(slap_state="destroy_requested")
    self.tic()

    with TemporaryAlarmScript(self.portal, 'InstanceTree_createUpgradeDecision'):
      self.portal.portal_alarms.slapos_pdm_instance_tree_create_upgrade_decision.\
        activeSense()
      self.tic()

    self.assertNotEqual('Visited by InstanceTree_createUpgradeDecision',
      instance_tree.workflow_history['edit_workflow'][-1]['comment'])

  def test_alarm_destroy_unused_software_installation(self):
    software_installation = self.portal.software_installation_module.newContent(
      portal_type='Software Installation'
    )
    self.portal.portal_workflow._jumpToStateFor(software_installation,
                                                'start_requested')
    self.portal.portal_workflow._jumpToStateFor(software_installation,
                                                'validated')
    self.tic()

    with TemporaryAlarmScript(self.portal, 'SoftwareInstallation_destroyIfUnused'):
      self.portal.portal_alarms.slapos_pdm_destroy_unused_software_installation.\
        activeSense()
      self.tic()

    self.assertEqual('Visited by SoftwareInstallation_destroyIfUnused',
      software_installation.workflow_history['edit_workflow'][-1]['comment'])